assert(BananaDKP, "This module requires BananaDKP.")

local BananaDKP = BananaDKP

local lastDkpCacheUpdate = {}

local counterRunning = nil

local FLASK_MULTIPLIER = 0.4
local ELIXIR_MULTIPLIER = 0.3
local FOOD_MULTIPLIER = 0.1

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = setmetatable({}, {__index =
	function(self, key)
		self[key] = key
		return key
	end
})

local flasks = {
	["Flask of Arcane Fortification"] = true,
	["Chromatic Resistance"] = true,
	["Distilled Wisdom"] = true,
	["Flask of Fortification"] = true,
	["Flask of Mighty Restoration"] = true,
	["Flask of Relentless Assault"] = true,
	["Flask of Shadow Fortification"] = true,
	["Supreme Power"] = true,
	["Flask of Pure Death"] = true,
	["Flask of the Titans"] = true,
	["Supreme Power of Shattrath"] = true,
	["Fortification of Shattrath"] = true,
	["Mighty Restoration of Shattrath"] = true,
	["Relentless Assault of Shattrath"] = true,
	["Unstable Flask of the Bandit"] = true,
	["Unstable Flask of the Beast"] = true,
	["Unstable Flask of the Elder"] = true,
	["Unstable Flask of the Sorcerer"] = true,
	["Unstable Flask of the Soldier"] = true,
	["Unstable Flask of the Physician"] = true,
}

local elixirs = {
	["Healing Power"] = true,
	["Elixir of Draenic Wisdom"] = true,
	["Elixir of Brute Force"] = true,
	["Fel Strength Elixir"] = true,
	["Empowerment"] = true,
	["Greater Mana Regeneration"] = true,
	["Major Shadow Power"] = true,
	["Major Armor"] = true,
	["Elixir of Ironskin"] = true,
	["Major Agility"] = true,
	["Earthen Elixir"] = true,
	["Major Frost Power"] = true,
	["Elixir of Mastery"] = true,
	["Elixir of Major Fortitude"] = true,
	["Major Strength"] = true,
	["Adept's Elixir"] = true,
	["Onslaught Elixir"] = true,
	["Greater Arcane Elixir"] = true,
	["Elixir of the Mongoose"] = true,
	["Greater Firepower"] = true,
	["Shadow Power"] = true,
	["Elixir of the Giants"] = true,
	["Arcane Elixir"] = true,
}

local lootDialogOpen = nil
local lootQueue = {}

local mode = nil
local host = nil

local halfHour = 60 * 30
local oneHour = halfHour * 2

local paused = nil
local counterRunning = nil

local db = nil

local _G = getfenv(0)

Symphony = BananaDKP:NewModule("Symphony")

Symphony.defaultDatabase = {
	databaseVersion = 2, -- Database version.
	items = {}, -- Key: Player, Value: Table of item ids looted.
	dkp = {},
	cachedDkp = {}, -- Only used by clients.
	hourTickerSavePoint = 0,
}

local function disable()
	return not UnitInRaid("player") or counterRunning
end
Symphony.options = {
	autoHeader = {
		type = "header",
		name = L["Automatic"],
		order = 100,
	},
	start = {
		type = "execute",
		name = L["Start"],
		desc = L["Start the automatic hour counter."],
		func = "StartCounter",
		order = 101,
		disabled = disable,
	},
	pause = {
		type = "toggle",
		name = L["Pause"],
		desc = L["Pause the automatic hour counter."],
		get = function() return paused and true or false end,
		set = "PauseCounter",
		order = 102,
		disabled = disable,
	},
	stop = {
		type = "execute",
		name = L["Stop"],
		desc = L["Stop the automatic hour counter."],
		func = "StopCounter",
		disabled = disable,
		order = 103,
	},
	spacer = {
		type = "header",
		name = " ",
		order = 150,
	},
	manualHeader = {
		type = "header",
		name = L["Manual"],
		order = 200,
	},
	plusone = {
		type = "text",
		name = L["Increment"],
		desc = L["Increment the hour counter for a player or the entire raid+sitout list (*) by one."],
		usage = L["<player or *>"],
		get = false,
		set = "IncrementHourCounter",
		order = 201,
	},
	minusone = {
		type = "text",
		name = L["Decrement"],
		desc = L["Decrement the hour counter for a player or the entire raid+sitout list (*) by one."],
		usage = L["<player or *>"],
		get = false,
		set = "DecrementHourCounter",
		order = 202,
	},
}

-- player is the person who looted, input is the text in the input box
local function addItemFromDialog(input, player, itemLink)
	local looter = player
	if type(input) == "string" then
		if UnitExists(input) then
			looter = input:gsub("^%l", string.upper)
		else
			BananaDKP:Print(L["%q is not in your group."]:format(input))
			return false
		end
	end
	Symphony:AddItemToItemTable(looter, itemLink)
	return true
end

local function itemWhisper(event, player)
	if mode == 2 then
		if not host then return end
		return L["I'm not the DKP host, whisper %s instead."]:format(host)
	end
	player = player:gsub("^%l", string.upper)
	if db.items[player] then
		local items = {}
		for i, v in ipairs(db.items[player]) do
			local itemId = select(3, v:find("item:(%d+):"))
			local name, link = GetItemInfo(itemId)
			if link then
				table.insert(items, link)
			elseif name then
				table.insert(items, name)
			else
				table.insert(items, (select(3, v:find("%[(.*)%]"))))
			end
		end
		return { L["Items looted by %s:"]:format(player), items, L["(%d items)"]:format(#items) }
	else
		return L["No items looted by %s"]:format(player)
	end
end

function Symphony:OnEnable()
	local popup = _G.StaticPopupDialogs
	if type(popup) ~= "table" then
		popup = {}
	end
	--[[ Event order for the popup
		OnShow
		EscapePressed
		Hide

		Show
		Cancel
		Hide

		Show
		Accept
		Hide

		Show
		EnterPressed
		Hide
	]]
	if type(popup["SymphonyItemQuery"]) ~= "table" then
		popup["SymphonyItemQuery"] = {
			button1 = L["Add"],
			button2 = L["Ignore"],
			sound = "AuctionWindowClose",
			whileDead = 1,
			timeout = 0,
			OnShow = function()
				_G[this:GetName().."EditBox"]:SetText("")
			end,
			OnHide = function()
				_G[this:GetName().."EditBox"]:SetText("")
			end,
			OnCancel = function()
				Symphony:IgnoreItem()
			end,
			EditBoxOnEscapePressed = function()
				this:GetParent():Hide()
				Symphony:IgnoreItem()
			end,
			EditBoxOnEnterPressed = function(data)
				local input = _G[this:GetParent():GetName().."EditBox"]:GetText()
				if addItemFromDialog(input, unpack(data)) then
					this:GetParent():Hide()
				end
			end,
			OnAccept = function(data)
				local input = _G[this:GetParent():GetName().."EditBox"]:GetText()
				if not addItemFromDialog(input, unpack(data)) then
					return true
				end
			end,
			hasEditBox = 1,
		}
	end

	db = self.db.profile

	if not db.databaseVersion or db.databaseVersion < 3 then
		db.databaseVersion = 3

		local newItemsTable = {}
		for player, items in pairs(db.items) do
			local newPlayer = player:gsub("^%l", string.upper)
			newItemsTable[newPlayer] = items
		end
		db.items = newItemsTable

		local newDKPTable = {}
		for player, dkp in pairs(db.dkp) do
			local newPlayer = player:gsub("^%l", string.upper)
			newDKPTable[newPlayer] = dkp
		end
		db.dkp = newDKPTable
	end

	BananaDKP:RegisterWhisperHandler("items", itemWhisper, L["bdkp items <player>; list of the items the player has looted"])

	self:RegisterEvent("RAID_ROSTER_UPDATE")

	mode = BananaDKP:GetMode()
	host = BananaDKP:GetHost()

	self:RegisterEvent("BananaDKP_HostChanged")
	self:RegisterEvent("BananaDKP_ModeChanged")

	self:RegisterEvent("CHAT_MSG_ADDON")
end

local function addonMsg(prefix, message, distribution, target)
	if ChatThrottleLib then
		ChatThrottleLib:SendAddonMessage("BULK", prefix, message, distribution, target)
	else
		SendAddonMessage(prefix, message, distribution, target)
	end
end

local grouped = nil
function Symphony:RAID_ROSTER_UPDATE()
	if not grouped and GetNumRaidMembers() > 0 then
		grouped = true
	elseif grouped and GetNumRaidMembers() == 0 then
		self:StopCounter()
		BananaDKP:ClearSitoutList()
		grouped = nil
	end
end

function Symphony:BananaDKP_ModeChanged()
	mode = BananaDKP:GetMode()
	if counterRunning then
		self:StopCounter()
	end
	for k in pairs(db.cachedDkp) do
		db.cachedDkp[k] = nil
	end
	for k in pairs(lastDkpCacheUpdate) do
		lastDkpCacheUpdate[k] = nil
	end

	self:TriggerEvent("BananaDKP_DKPUpdated")
end

function Symphony:BananaDKP_HostChanged()
	host = BananaDKP:GetHost()
	if counterRunning then
		self:StopCounter()
	end
	self:TriggerEvent("BananaDKP_DKPUpdated")
end

local function addItem(player, itemLink)
	if not db.items[player] then db.items[player] = {} end
	table.insert(db.items[player], itemLink)
	-- DKP -25%
	local dkp = db.dkp[player]
	local price = 0
	if not dkp then
		dkp = 0
	else
		price = dkp * 25 / 100
	end
	db.dkp[player] = dkp - price

	return price
end

local function incrementByOne(input)
	if db.dkp[input] then
		db.dkp[input] = db.dkp[input] + 1
	else
		db.dkp[input] = 1
	end
end

local function decrementByOne(input)
	if db.dkp[input] then
		db.dkp[input] = db.dkp[input] - 1
	end
end

local function inGroup(player)
	return UnitInRaid(player) or UnitInParty(player) and true or nil
end

-- Max addon message length is 255, and the max name length for WoW is 12.
-- We need 14 chars per name for name + space + separator, and DKP values
-- can be quite long, so we'll allocate 10 chars per return value.
-- So if we query for 10 players at a time, we get 100 + 140 = 240 which leaves
-- 14 chars for the prefix and any random overhead.

local dkpQueryList = {}
local function actuallyQueryDKP()
	if not host then return end
	addonMsg("BDKPR", table.concat(dkpQueryList, "#"), "WHISPER", host)
	for k in pairs(dkpQueryList) do dkpQueryList[k] = nil end
end

local function queueDkpQuery(player)
	table.insert(dkpQueryList, player)
	if #dkpQueryList < 10 then
		Symphony:ScheduleEvent("SymphonyQueryDKP", actuallyQueryDKP, 1)
	else
		Symphony:CancelScheduledEvent("SymphonyQueryDKP")
		actuallyQueryDKP()
	end
end

local function replyDkp(sender, ...)
	local r = {}
	for i = 1, select("#", ...) do
		local player = select(i, ...)
		local dkp = Symphony:GetDKP(player)
		if dkp then
			table.insert(r, player)
			table.insert(r, tostring(dkp))
		end
	end
	addonMsg("BDKP", table.concat(r, "#"), "WHISPER", sender)
	for k in pairs(r) do r[k] = nil end
end

local function parseDkpReply(...)
	for i = 1, select("#", ...), 2 do
		local player, dkp = select(i, ...)
		db.cachedDkp[player] = tonumber(dkp)
	end
end

function Symphony:CHAT_MSG_ADDON(prefix, message, distribution, sender)
	if not sender or not prefix or not message then return end
	if prefix == "BDKPI" and mode == 1 then
		if inGroup(sender) then return end -- Ignore anything from our own group.
		local player, itemLink = select(3, message:find("(%S+) (.*)"))
		if not player and itemLink then return end
		local price = addItem(player, itemLink)
		BananaDKP:Print(L["%s says %s looted %s, %d has been deducted from his DKP."]:format(sender, player, itemLink, price or 0))
		self:TriggerEvent("BananaDKP_DKPUpdated", sender)
	elseif prefix == "BDKPH" and mode == 1 then
		if inGroup(sender) then return end -- Ignore anything from our own group.

		local input, action = select(3, message:find("(%S+) (.*)"))
		if action == "+" then
			incrementByOne(input)
			BananaDKP:Print(L["%s manually incremented the DKP for %s by +1."]:format(sender, input))
		elseif action == "-" then
			decrementByOne(input)
			BananaDKP:Print(L["%s manually decremented the DKP for %s by -1."]:format(sender, input))
		elseif tonumber(action) then
			action = tonumber(action)
			if not db.dkp[input] then
				db.dkp[input] = action
			else
				db.dkp[input] = db.dkp[input] + action
			end
		end
		self:TriggerEvent("BananaDKP_DKPUpdated", sender)
	elseif prefix == "BDKP" and mode == 2 and sender == host then
		if message:find("#") then
			parseDkpReply(strsplit("#", message))
		else
			local player, dkp = select(3, message:find("(%S+) (.*)"))
			db.cachedDkp[player] = tonumber(dkp)
		end
		self:TriggerEvent("BananaDKP_DKPUpdated")
	elseif prefix == "BDKPR" and mode == 1 then
		if message:find("#") then
			replyDkp(sender, strsplit("#", message))
		else
			local dkp = self:GetDKP(message)
			addonMsg("BDKP", message .. " " .. tostring(dkp), "WHISPER", sender)
		end
		self:TriggerEvent("BananaDKP_DKPUpdated", sender)
	end
end

-----------------------------------------------------------------------
-- Ticker handling

-- Every minute we save our current timer, so that if we get disconnected or
-- something, we know where to start from next time.
function Symphony:RepeatedTimeDBStore()
	if not counterRunning then return end
	local timeLeft
	if paused then
		timeLeft = paused
	else
		timeLeft = select(2, self:IsEventScheduled("Ticker"))
	end
	db.hourTickerSavePoint = timeLeft
	self:ScheduleEvent("TickerStore", self.RepeatedTimeDBStore, 60, self)
end

local function startCounter()
	local time = 0
	if db.hourTickerSavePoint > 0 then
		time = db.hourTickerSavePoint
	else
		time = halfHour
	end
	Symphony:ScheduleEvent("Ticker", Symphony.IncrementHourCounter, time, Symphony)
	counterRunning = true
	Symphony:RepeatedTimeDBStore()
	local minutes = math.floor(time / 60)
	BananaDKP:Print(L["Automatic hour counting has started, the first increment will be in ~%d minutes."]:format(minutes))
	BananaDKP:Chat(L["BDKP hour counter started, whisper me 'sitout' if you are not in the raid but available on standby."], "GUILD")
end

function Symphony:StartCounter()
	assert(not counterRunning, "Counter is already running.")
	startCounter()
end

function Symphony:PauseCounter(input)
	assert(counterRunning, "Can't pause something that is not running.")
	if input then
		paused = select(2, self:IsEventScheduled("Ticker"))
		self:CancelScheduledEvent("Ticker")
		local minutes = math.floor(paused / 60)
		BananaDKP:Print(L["Automatic hour counting paused, ~%d minutes to next update."]:format(minutes))
	else
		self:ScheduleEvent("Ticker", self.IncrementHourCounter, paused, self)
		paused = nil
		BananaDKP:Print(L["Automatic hour counting resumed."])
	end
end

function Symphony:StopCounter()
	if not counterRunning then return end
	local timeLeft = select(2, self:IsEventScheduled("Ticker"))
	if not timeLeft then timeLeft = paused end
	self:CancelScheduledEvent("Ticker")
	self:CancelScheduledEvent("TickerStore")
	counterRunning = nil
	timeLeft = math.floor(timeLeft / 60)
	BananaDKP:Print(L["Automatic hour counting stopped, ~%d minutes to next update."]:format(timeLeft or -1))
	paused = nil
	db.hourTickerSavePoint = 0
end

-----------------------------------------------------------------------
-- Item handling

function Symphony:ItemLooted(player, itemLink, itemRarity)
	if mode == 2 and host and inGroup(host) then return end

	if itemRarity < 4 then return end
	if lootDialogOpen then
		table.insert(lootQueue, {player, itemLink, itemRarity})
	else
		lootDialogOpen = true
		StaticPopupDialogs["SymphonyItemQuery"].text = L["%s looted %s, click Add to add it to his loot table, or enter the name of someone else who should pay for it in the text box below."]:format(player, itemLink)
		local popup = StaticPopup_Show("SymphonyItemQuery")
		if popup then
			popup.data = {player, itemLink}
		end
	end
end

function Symphony:AddItemToItemTable(player, itemLink)
	lootDialogOpen = nil

	-- mode 1 = server, 2 = client
	if mode == 1 then
		local price = addItem(player, itemLink)
		BananaDKP:Print(L["%s looted %s, %d has been deducted from his DKP."]:format(player, itemLink, price or 0))
	elseif mode == 2 then
		if not host then return end
		addonMsg("BDKPI", player .. " " .. itemLink, "WHISPER", host)
		BananaDKP:Print(L["%s looted %s, it has been added to his item table."]:format(player, itemLink))
	end

	if #lootQueue > 0 then
		local n = table.remove(lootQueue, 1)
		self:ScheduleEvent("BDKPSItem", self.ItemLooted, 1, self, n[1], n[2], n[3])
	else
		self:TriggerEvent("BananaDKP_DKPUpdated")
	end
end
function Symphony:IgnoreItem()
	lootDialogOpen = nil

	if #lootQueue > 0 then
		local n = table.remove(lootQueue, 1)
		self:ScheduleEvent("BDKPSItem", self.ItemLooted, 1, self, n[1], n[2], n[3])
	end
end

-----------------------------------------------------------------------
-- Hour handling

function Symphony:IncrementHourCounter(input)
	if type(input) == "string" and not input:find("*") then
		input = input:trim():gsub("^%l", string.upper)
		if mode == 1 then
			incrementByOne(input)
		elseif mode == 2 then
			if not host then return end
			addonMsg("BDKPH", input .. " +", "WHISPER", host)
		end
		BananaDKP:Print(L["The hour counter has been incremented by 1 for %s."]:format(input))
	else
		if type(input) == "string" and input ~= "*" then
			error("Can't increment the hour counter for "..input..".")
			return
		end
		local dkpTable = {}
		for i = 1, GetNumRaidMembers() do
			local n = GetRaidRosterInfo(i)
			local multiplier = 1
			local buffIndex = 1
			local buff = UnitBuff(n, buffIndex)
			while buff do
				if flasks[buff] then
					multiplier = multiplier + FLASK_MULTIPLIER
				elseif buff == "Well Fed" then
					multiplier = multiplier + FOOD_MULTIPLIER
				elseif elixirs[buff] then
					multiplier = multiplier + ELIXIR_MULTIPLIER
				end
				buffIndex = buffIndex + 1
				buff = UnitBuff(n, buffIndex)
			end
			dkpTable[n] = 1 * multiplier
		end
		for k, v in pairs(BananaDKP.db.profile.sitouts) do
			if not dkpTable[k] then dkpTable[k] = 1 end
		end
		if mode == 1 then
			for k, v in pairs(dkpTable) do
				if not db.dkp[k] then
					db.dkp[k] = v
				else
					db.dkp[k] = db.dkp[k] + v
				end
			end
		elseif mode == 2 then
			if not host then return end
			for k, v in pairs(dkpTable) do
				addonMsg("BDKPH", k .. " " .. tostring(v), "WHISPER", host)
			end
		end
		BananaDKP:Print(L["The hour counter has been incremented by 1 for all players in the raid and on the sitout list."])
		if type(input) ~= "string" then
			self:ScheduleEvent("Ticker", self.IncrementHourCounter, oneHour, self)
		end
	end

	self:TriggerEvent("BananaDKP_DKPUpdated")
end

function Symphony:DecrementHourCounter(input)
	if type(input) == "string" and not input:find("*") then
		input = input:trim():gsub("^%l", string.upper)
		if mode == 1 then
			decrementByOne(input)
		elseif mode == 2 then
			if not host then return end
			addonMsg("BDKPH", input .. " -", "WHISPER", host)
		end
		BananaDKP:Print(L["The hour counter has been decremented by 1 for %s."]:format(input))
	else
		if type(input) == "string" and input ~= "*" then
			error("Can't decrement the hour counter for "..input..".")
			return
		end
		local dkpTable = {}
		for i = 1, GetNumRaidMembers() do
			local n = GetRaidRosterInfo(i)
			dkpTable[n] = true
		end
		for k, v in pairs(BananaDKP.db.profile.sitouts) do
			dbkTable[k] = true
		end
		if mode == 1 then
			for k, v in pairs(dkpTable) do
				if db.dkp[k] then
					db.dkp[k] = db.dkp[k] - 1
				end
			end
		elseif mode == 2 then
			if not host then return end
			for k, v in pairs(dkpTable) do
				addonMsg("BDKPH", k .. " -", "WHISPER", host)
			end
		end
		BananaDKP:Print(L["The hour counter has been decremented by 1 for all players in the raid and on the sitout list."])
		if type(input) ~= "string" then
			self:ScheduleEvent("Ticker", self.IncrementHourCounter, oneHour, self)
		end
	end

	self:TriggerEvent("BananaDKP_DKPUpdated")
end

-----------------------------------------------------------------------
-- DKP handling

-- DKP Formula:
-- #hoursRaided / (1 + #itemsLooted)
--
-- Pretty simple, I know :P That's what my guild wanted, so that's what I did.

-- Thanks, Wobin.
local function round(n,shift)
	shift = 10^shift
	return math.floor((n*shift)+0.5)/shift
end

function Symphony:GetDKP(name)
	if not name then return end
	if mode == 1 then
		if not db.dkp[name] then return end
		return round(db.dkp[name], 2)
	elseif mode == 2 then
		if not lastDkpCacheUpdate[name] or (GetTime() > (lastDkpCacheUpdate[name] + 60)) then
			lastDkpCacheUpdate[name] = GetTime()
			if not host then return end
			queueDkpQuery(name)
		end
		if not db.cachedDkp[name] then return end
		return round(db.cachedDkp[name], 2)
	end
end

